package mn;
import fonction.*;
import matrice.*;
import java.lang.Math;

/**
 * Classe de calcul du zro d'une fonction par diverses mthodes :
 *
 *  -largissement de l'intervalle
 *  -Dichotomie
 *  -Algorithme de Newton-Raphson
 *  -Algorithme de Newton multidimensionnel
 *
 * @see fonction.ChoixDeFonction ChoixDeFonction
 *
 * @author Benoit CARRY (version du 17/01/2006)
 * <p>Mathieu HAVEL, Pierre INIZAN (version du 10/01/2007)
 * @version 10/01/2007
 * @since Projet Trois Corps 2006
 */
public class DetermineZero {


   /**
    * Constructeur de la classe DetermineZero
    *
    */
  public DetermineZero() {}



    /**
     * Mthode de dtermination du zro d'une fonction  partir de celle-ci
     * Algorithme de dichotomie
     *
     *  @param intervalle Intervalle initial comprenant le zro de la fonction 
     *  @param precision Prcision souhaite
     *  @param paramIn Tableau de paramtres dont dpend la fonction dont on cherche le zro
     *  @param aMinimiser Fonction dont on cherche le zro
     *
     *  @return L'intervalle contenant le zro de la fonction,  la prcision souhaite
     *  @since Projet Trois Corps 2006
     */
  public double[] dichotomie(double[] intervalle, double precision, double[] paramIn, ChoixDeFonction aMinimiser) {

   //-Intermediaires de calcul pour effectuer la dichotomie
    double x1, x2, xtest;              //Abscisses des bornes
    double y1, y2, ytest;              //Ordonnees associees
    double[] resultat=new double[2];   //Couple d'abscisse des bornes


   //-Mise en place des couples (x,y) en debut et fin d'intervalle
    x1=intervalle[0];
     y1=aMinimiser.fonction( x1, paramIn);
    x2=intervalle[1];
     y2=aMinimiser.fonction( x2, paramIn);


   //-Dichotomie jusqu'a la precision souhaitee
    while ( Math.abs(x2-x1) >= precision && y1*y2 < 0 ) {

     //-On calcule la valeur mediane
      xtest = 0.5*( x1+x2 );
      ytest = aMinimiser.fonction( xtest, paramIn);

     //-Reduction de l'intervalle a moitie
      if ( y1*ytest<0 ) {
        x2 = xtest;
        y2 = aMinimiser.fonction( x2, paramIn);
      }
      else {
        x1 = xtest;
        y1 = aMinimiser.fonction( x1, paramIn);
      }

    }
  
   //-Affectation des bornes a un vecteur (pour la sortie)
    resultat[0] = x1;
    resultat[1] = x2;

    return resultat;

  }







    /**
     * Mthode de dtermination du zro d'une fonction  partir de celle-ci et de sa derive
     * Algorithme de Newton Raphson
     *
     *  @param position Position initiale proche du zro suppos
     *  @param precision Prcision souhaite
     *  @param paramIn Tableau de paramtres dont dpandent la fonction dont on cherche le zro et sa drive
     *  @param aMinimiser Fonction dont on cherche le zro
     *  @param deriveeDeMini Sa derive
     *
     *  @return Renvoie une approximation du zro de la fonction  la precision souhaite
     *  @since Projet Trois Corps 2006
     */
  public double newtonRaphson(double position, double precision, double[] paramIn, ChoixDeFonction aMinimiser, ChoixDeFonction deriveeDeMini) {

   //-Intermedaires de calcul
    double x,y,dy;                //Position de test, ordonnee et pente associees
    double ecart=precision+1e-6;  //Deplacement de la position test entre 2 pas
    double coeffDeSousRelax=1;    //Limitation du NR pour eviter les cycles
    int nbIter=0;                 //Nombre d'iterations

   //-Debugging
//    System.out.println("Debut de NewtonRaphson");
   //-Debugging

   //-Positions initiales
    x = position;
    y = aMinimiser.fonction( x, paramIn);

   //-Boucle jusqu'a la precision souhaitee (en x et en y)
    while ( Math.abs(ecart) >= precision || Math.abs(y) >= precision ) {

     //-Valeur de la fonction et de sa derivee au point test
      y  = aMinimiser.fonction( x, paramIn);
      dy = deriveeDeMini.fonction( x, paramIn);

     //-Calcul de la nouvelle position test
      ecart = y/dy;
      x = x - coeffDeSousRelax*ecart;

      nbIter++;
// -Debugging
//    System.out.println("Nombre d'iterations : "+nbIter+" x = "+x+" y = "+y+" dy = "+dy+" ecart = "+ecart);
//   -Debugging

      }

//   -Debugging
//    System.out.println("Nombre d'iterations : "+nbIter);
//   -Debugging

    return x;

  }


//  modif_b1_002
//  ajoute le : 07/11/06
//  modifie le : 07/01/07

  /**
   * Algorithme de Newton multidimensionnel : 
   *<p>
   * Mthode de dtermination du zro d'une fonction de R_n dans R_nbe  partir de celle-ci et de sa matrice jacobienne (matrice  nbe lignes et n colonnes)
   *
   *  @param nbe Dimension du systme (nombre d'quations)
   *  @param eps Prcision souhaite
   *  @param relax Paramtre de relaxation
   *  @param position Position du vecteur d'essai, de taille n
   *  @param fonc_res Fonction que l'on cherche  annuler
   *  @param par1 Paramtres de fonc_res (constantes intervenant dans sa structure)
   *  @param jac_fonc Matrice jacobienne de la fonction
   *  @param par2 Paramtres de jac_fonc (constantes intervenant dans sa structure)
   *  @return Renvoie une approximation du zero de la fonction  la precision souhaite
   *  @since Projet Trois Corps 2007
   */

  public double[] newton_multidim(int nbe, double eps, double relax, double[] position, ChoixDeFonction fonc_res, double[] par1, ChoixDeFonction jac_fonc, double[] par2) {

    int i,j,k;
    int n = position.length; //nombre de variables du systeme
    int dim = nbe*n; // nombre total de coefficients de la matrice jacobienne
    double errx, erry; // erreurs : erry = norme de la fonction au point courant, errx = norme de l'ecart entre deux positions successives 
    double[] pos = new double[n]; //valeur de la position a chaque iteration
    double[] jaco = new double[dim]; // tableau representant la matrice jacobienne J(x)
    double[] y = new double[nbe]; // valeur de la fonction  
    double[] Dx = new double[n]; // tel que jac_fonc(pos). Dx = fonc_res(pos)    

    errx = 1.;
    erry = 1.;

    for(i=0;i<n;i++){
      pos[i] = position[i];
    }

    y = fonc_res.fonction_multi(pos,par1,nbe); // valeur de la fonction au point pos

    int n_max = 10000;
    int ll = 0;  

    while(((errx > eps) || (erry > eps)) && (ll < n_max)){ // condition d'arret : position stable (errx <eps) ET fonction nulle en cette position (erry < eps)

      jaco = jac_fonc.fonction_multi(pos,par2,nbe); // valeur de la matrice jacobienne de la fonction au point pos
      Matrix jacoM = new Matrix(jaco,nbe); // conversion (tableau -> Matrice) de la matrice jacobienne colonne par colonne
      Matrix yM = new Matrix(y,nbe); // conversion (tableau -> Matrice) de la valeur de la fonction en pos
      Matrix DxM = new Matrix(n,1); // deplacement de la position (direction opposee au gradient)

      DxM = jacoM.solve(yM); // resolution de jacoM.DxM = yM par decomposition LU

      Dx = DxM.getColumnPackedCopy(); // conversion (Matrice -> tableau) de la direction  colonne par colonne

      //nouvelle position
      for(k=0;k<n;k++){
        pos[k] = pos[k] - relax*Dx[k];
      }

      //calcul de errx
      errx = 0; 
      for(i=0;i<n;i++){
        errx = errx + Math.abs(Dx[i]);  
      } 

      y = fonc_res.fonction_multi(pos,par1,nbe); // valeur de la fonction au nouveau point courant

      //calcul de erry
      erry = 0;
      for(j=0;j<nbe;j++){
        erry = erry + Math.abs(y[j]);  
      }

      ll++;

    }

    for(k=0;k<n;k++){
        pos[k] = pos[k] + relax*Dx[k] - Dx[k];
      } 
    //System.out.println("errx = "+errx);
    //System.out.println("erry = "+erry);
    //System.out.println("ll = "+ll);

    return pos; // valeur de la position ou la fonction s'annule
  }






    /**
     * Mthode de vrification de la prsence d'un zro dans l'intervalle
     *
     *  @param intervalle Intervalle initial comprenant le zro de la fonction 
     *  @param paramIn Tableau de paramtres dont dpend la fonction dont on cherche le zro
     *  @param aMinimiser Fonction dont on cherche le zero
     *
     *  @return Renvoie un intervalle contenant obligatoirement le zro de la fonction
     *  @since Projet Trois Corps 2006 
     */
  public double[] verifZero(double[] intervalle, double[] paramIn, ChoixDeFonction aMinimiser) {


   //-Intermediaires de calcul pour effectuer la verification
    double x1, x2;                      //Abscisses des bornes
    double y1, y2;                      //Ordonnees associees
    double[] resultat = new double[2];  //Couple d'abscisse des bornes
    double facteur=0.15;                //Facteur d'augmentation de l'intervalle


   //-Mise en place des couples (x,y) en debut et fin d'intervalle
    x1=intervalle[0];
     y1=aMinimiser.fonction( x1, paramIn);
    x2=intervalle[1];
     y2=aMinimiser.fonction( x2, paramIn);


   //-Debugging
    //if (y1*y2 <0) {
    //  System.out.println("Valeur dans l'intervalle propose...");
    //}
   //-Debugging


   //-Augmentation de la taille de l'intervalle jusqu'a contenir le zero
    while ( y1*y2 >= 0 ) {

     //-Augmente la taille de l'intervalle 
      x1 = x1 - (x2-x1)*facteur;
      x2 = x2 + (x2-x1)*facteur;
   
      y1=aMinimiser.fonction( x1, paramIn);
      y2=aMinimiser.fonction( x2, paramIn);

    }

  
   //Affectation des bornes a un vecteur (pour la sortie)
    resultat[0] = x1;
    resultat[1] = x2;


   //-Debugging
    //System.out.println("L'intervalle vaut : [ "+x1+" , "+x2+" ]");
   //-Debugging

    return resultat;

  }





}
